# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
sys.path.insert(0, os.path.abspath('../arcus'))

# One annoyance is that the magic methods ignored
# while often this makes sense. Sometimes it does not. 
# To override this , we are adding the following
def skip(app, what, name, obj,would_skip, options):
    if name in ( '__init__',):
        return False
    return would_skip
def setup(app):
    app.connect('autodoc-skip-member', skip)

# -- Project information -----------------------------------------------------

project = 'arcus.ml'
copyright = '2021, Arcus Machine Learning'
author = 'Arcus Machine Learning'


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon'
]

# Configuration of napoleon

# Napoleon settings
napoleon_google_docstring = True
napoleon_numpy_docstring = False
napoleon_include_init_with_doc = False
napoleon_include_private_with_doc = False
napoleon_include_special_with_doc = False
napoleon_use_admonition_for_examples = False
napoleon_use_admonition_for_notes = False
napoleon_use_admonition_for_references = False
napoleon_use_ivar = False
napoleon_use_param = True
napoleon_use_rtype = True
napoleon_preprocess_types = False
napoleon_type_aliases = None
napoleon_attr_annotations = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = 'en'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']


# -- Extension configuration -------------------------------------------------

# -- Options for todo extension ----------------------------------------------

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


import sys
import os
import pytest
import pandas as pd 
import arcus.ml.images.io as ami
import arcus.ml.images.conversion as conv
import logging
import numpy as np 

image_path = 'tests/resources/images/lungs'

def test_black_white_conversion():
    image_list = ami.load_images(image_path)
    ls = conv.to_blackwhite(image_list)
    pixel_value = ls[0][0][0]
    assert pixel_value == 0 or pixel_value == 1
    assert len(ls[0].shape)==2

def test_black_white_conversion_double_arrays():
    image_list = ami.load_images(image_path)
    test_list = list()
    for img in image_list:
        test_list.append(img / 255)
    ls = conv.to_blackwhite(np.array(test_list))
    pixel_value = ls[0][0][0]
    assert pixel_value == 0 or pixel_value == 1

def test_crop_image():
    image = ami.load_images(image_path, max_images=1, image_size=50)[0]
    crop = conv.crop(image, 10, 20, 20, 10)
    assert crop.shape == (10, 20, 3)

def test_black_white_conversion_3dshape():
    image_list = ami.load_images(image_path)
    ls = conv.to_blackwhite(image_list, keep_3d_shape=True)
    pixel_value = ls[0][0][0]
    assert pixel_value == 0 or pixel_value == 1
    assert len(ls[0].shape)==3

def test_black_white_conversion_2x3dshape():
    lung = ami.load_images(image_path, convert_to_grey=True, image_size = 30, keep_3d_shape=True)[0]
    assert lung.shape == (30, 30, 1)
    lung = conv.to_blackwhite([lung], keep_3d_shape=True, threshold=200)
    assert lung[0].shape == (30, 30, 1)


import sys
import os
import pytest
import pandas as pd 
import arcus.ml.dataframes as adf
import logging
from unittest.mock import patch 
from collections import Counter

logging.basicConfig(level=logging.DEBUG)
mylogger = logging.getLogger()

# initialize list of lists 
categorical_data = [['BE', 10], ['FR', 15], ['BE', 14], ['UK', 14], ['SE', 14], ['SE', 14]] 


def setup_module(module):
    ''' Setup for the entire module '''
    # Do the actual setup stuff here
    pass

def setup_function(func):
    ''' Setup for test functions '''


def test_shuffle():
    # initialize list of lists 
    data = [['tom', 10], ['nick', 15], ['juli', 14]] 
    
    # Create the pandas DataFrame 
    test_df = pd.DataFrame(data, columns = ['Name', 'Age']) 
    
    # print dataframe. 
    shuffled_df = adf.shuffle(test_df)
    assert len(shuffled_df)==len(test_df)
    assert shuffled_df[shuffled_df.Name == 'nick'].iloc[0].Age == 15

def test_one_hot_encoding_default():
    # Create the pandas DataFrame 
    test_df = pd.DataFrame(categorical_data, columns = ['Country', 'Infected']) 

    # One hot encode the Country column
    encoded_df = adf.one_hot_encode(test_df, 'Country')
    assert len(encoded_df.columns) == 5
    assert 'Country_BE' in encoded_df.columns
    assert len(encoded_df.loc[encoded_df['Country_BE'] == 1]) == 2
    assert len(encoded_df.loc[encoded_df['Country_FR'] == 0]) == 5


def test_one_hot_encoding_keepcolumn():
    # Create the pandas DataFrame 
    test_df = pd.DataFrame(categorical_data, columns = ['Country', 'Infected']) 

    # One hot encode the Country column
    encoded_df = adf.one_hot_encode(test_df, 'Country', drop_column=False)
    assert len(encoded_df.columns) == 6
    assert 'Country_SE' in encoded_df.columns
    assert len(encoded_df.loc[encoded_df['Country_UK'] == 1]) == 1


@patch("matplotlib.pyplot.show")
def test_one_hot_encoding_prefix(mock_show):
    # Create the pandas DataFrame 
    test_df = pd.DataFrame(categorical_data, columns = ['Country', 'Infected']) 
    encoded_df = adf.one_hot_encode(test_df, 'Country', prefix='cnt')

    # One hot encode the Country column
    assert len(encoded_df.columns) == 5
    assert 'cnt_BE' in encoded_df.columns
    assert len(encoded_df.loc[encoded_df['cnt_UK'] == 1]) == 1

@patch("matplotlib.pyplot.show")
def test_plot_features_selectedcols(mock_show):
    df = pd.read_csv('tests/resources/datasets/car-fuel.csv')
    _, _axes = adf.plot_features(df, column_names=['speed', 'temp_outside'])
    assert len(_axes.ravel()) == 2
    _ax = _axes[0]
    assert _ax.numCols == 2
    assert _ax.numRows == 1

@patch("matplotlib.pyplot.show")
def test_plot_features_onerow(mock_show):
    df = pd.read_csv('tests/resources/datasets/car-fuel.csv')
    _, _axes = adf.plot_features(df, column_names=['speed', 'temp_outside'])
    assert len(_axes.ravel()) == 2
    _ax = _axes[0]
    assert _ax.numCols == 2
    assert _ax.numRows == 1

@patch("matplotlib.pyplot.show")
def test_plot_features_grid_size(mock_show):
    df = pd.read_csv('tests/resources/datasets/car-fuel.csv')
    _, _axes = adf.plot_features(df, grid_shape=(2,3))
    assert len(_axes.ravel()) == 6
    _ax = _axes[0][0]
    assert _ax.numCols == 3
    assert _ax.numRows == 2

@patch("matplotlib.pyplot.show")
def test_plot_features_default(mock_show):
    df = pd.read_csv('tests/resources/datasets/car-fuel.csv')
    _, _axes = adf.plot_features(df)
    assert len(_axes.ravel()) == 5
    assert _axes[0].numCols == 5
    assert _axes[0].numRows == 1

def test_keep_numeric_features():
    test_df = pd.DataFrame(categorical_data, columns = ['Country', 'Infected']) 
    num_df = adf.keep_numeric_features(test_df)
    assert len(num_df.columns) == 1

def test_keep_numeric_features_onehotencoded():
    test_df = pd.DataFrame(categorical_data, columns = ['Country', 'Infected'])
    encoded_df = adf.one_hot_encode(test_df, 'Country', prefix='cnt') 
    num_df = adf.keep_numeric_features(encoded_df)
    assert len(num_df.columns) == 5

def test_keep_numeric_features_csv():
    df = pd.read_csv('tests/resources/datasets/car-fuel.csv')
    num_df = adf.keep_numeric_features(df)
    assert len(num_df.columns) == 5

def test_class_distribution_default():
    df = pd.read_csv('tests/resources/datasets/car-fuel.csv')
    cnt1 = Counter(df.gas_type)
    df = adf.distribute_class(df, 'gas_type')
    cnt2 = Counter(df.gas_type)
    # Smallest class size from first should be same as largest from new
    assert min(cnt1.values()) == max (cnt2.values())
    # Sizes of new df should be the same
    assert min(cnt2.values()) == max (cnt2.values())
    assert len(df) == min(cnt1.values()) * 2

def test_class_distribution_class_size():
    cs = 30
    df = pd.read_csv('tests/resources/datasets/car-fuel.csv')
    cnt1 = Counter(df.gas_type)
    df = adf.distribute_class(df, 'gas_type', cs)
    cnt2 = Counter(df.gas_type)
    # Sizes of new df should be the same
    assert min(cnt2.values()) == max (cnt2.values())
    assert len(df) == cs * 2

def test_class_distribution_noshuffle():
    cs = 30
    df = pd.read_csv('tests/resources/datasets/car-fuel.csv')
    cnt1 = Counter(df.gas_type)
    df = adf.distribute_class(df, 'gas_type', cs, False)
    cnt2 = Counter(df.iloc[0:cs].gas_type)
    assert len(cnt2.keys()) == 1


'''
The keras module provides additions to work and visualize Keras neural networks
'''

import tensorflow as tf
from tensorflow import keras
import logging
import os

_logger = logging.getLogger()

def enable_gpu() -> bool:
    '''
    Enables Keras to run on the GPU
    First it checks if there's a "physical device" available of type GPU
    If not, it will try to leverage plaidml for GPU enabling

    Returns: 
        bool: A boolean value indicating if the GPU enablement succeeded
    '''

    gpu_enabled = False
    physical_devices = tf.config.experimental.list_physical_devices('GPU')
    if(len(physical_devices)>0):
        tf.config.experimental.set_memory_growth(physical_devices[0], True)
        _logger.info('Physical GPU enabled for tensorflow')
        gpu_enabled = True
    else:
        _logger.info('No physical GPU found, trying plaidml now')
        try:
            import plaidml.keras
            gpu_enabled = True
            plaidml.keras.install_backend()
            os.environ["KERAS_BACKEND"] = "plaidml.keras.backend"
        except ImportError as e:
            _logger.warning('GPU could not be enabled, as plaidml package was not installed')
    
    return gpu_enabled



import setuptools 

def test_package_discovery():
    assert len(setuptools.find_packages('arcus')) > 0
    assert len(setuptools.find_packages('tests')) == 0
    assert len(setuptools.find_packages('samples')) == 0



from lazydocs import generate_docs
import pytest
import glob, os, os.path

def test_generate_apidocs():
    # Remove all files
    files_to_ignore = ['ml.images.md', 'ml.md', 'ml.evaluation.md', 'ml.neuralnetworks.md', 'ml.timeseries.md']
    sdk_docs_dir = 'docs/preview/sdk'
    filelist = glob.glob(os.path.join(sdk_docs_dir, "*.md"))
    for f in filelist:
        os.remove(f)
    generate_docs(['arcus'], output_path=sdk_docs_dir, overview_file="index.md", remove_package_prefix=True)
    for f in files_to_ignore:
        os.remove(os.path.join(sdk_docs_dir, f))
    filelist = glob.glob(os.path.join(sdk_docs_dir, "*.md"))
    for f in filelist:
        __fix_html_file(f)

def __fix_html_file(md_file:str):
    # Read in the file
    with open(md_file, 'r') as file :
        filedata = file.read()

    # Close the img tag
    filedata = filedata.replace('></a>', ' /></a>')
    # Remove the style of images
    filedata = filedata.replace('style="float:right;"', '')

    # Write the file out again
    with open(md_file, 'w') as file:
        file.write(filedata)


import sys
import os
import pytest
import pandas as pd 
import arcus.ml.images.explorer as exp
import arcus.ml.images.io as imgio
import arcus.ml.images.conversion as conv
import logging

image_path_1 = 'tests/resources/images/lungs'
image_path_2 = 'tests/resources/images/lungmasks'

def test_visualize_random_images():
    lungs = imgio.load_images(image_path_1)
    masks = imgio.load_images(image_path_2)

    exp.visualize({'lungs': lungs, 'masks': masks}, image_count = 6, randomize=True)

def test_visualize_first_images():
    lungs = imgio.load_images(image_path_1)
    masks = imgio.load_images(image_path_2)

    exp.visualize({'lungs': lungs, 'masks': masks}, image_count = 6, randomize=False)

def test_visualize_3d_shape_images():
    lungs = imgio.load_images(image_path_1, convert_to_grey=True, keep_3d_shape=True)
    masks = imgio.load_images(image_path_2, convert_to_grey=True, keep_3d_shape=True)

    exp.visualize({'lungs': lungs, 'masks': masks}, image_count = 6, randomize=True)


def test_visualize_oneset_images():
    lungs = imgio.load_images(image_path_1)

    exp.visualize({'lungs': lungs}, image_count = 6, randomize=True)

def test_show_3d_shape_image():
    lungs = imgio.load_images(image_path_1, convert_to_grey=True, keep_3d_shape=True)
#X = conversion.to_blackwhite(X, keep_3d_shape=True, threshold=200)
    exp.show_image(lungs[1])

def test_visualize_random_images_silent():
    lungs = imgio.load_images(image_path_1)
    masks = imgio.load_images(image_path_2)

    pp = exp.visualize({'lungs': lungs, 'masks': masks}, image_count = 6, randomize=True, silent_mode=True)
    assert pp is not None

def test_visualize_first_images_silent():
    lungs = imgio.load_images(image_path_1)
    masks = imgio.load_images(image_path_2)

    pp = exp.visualize({'lungs': lungs, 'masks': masks}, image_count = 6, randomize=False, silent_mode=True)
    assert pp is not None

def test_visualize_3d_shape_images_silent():
    lungs = imgio.load_images(image_path_1, convert_to_grey=True, keep_3d_shape=True)
    masks = imgio.load_images(image_path_2, convert_to_grey=True, keep_3d_shape=True)

    pp = exp.visualize({'lungs': lungs, 'masks': masks}, image_count = 6, randomize=True, silent_mode=True)
    assert pp is not None


def test_visualize_oneset_images_silent():
    lungs = imgio.load_images(image_path_1)

    pp = exp.visualize({'lungs': lungs}, image_count = 6, randomize=True, silent_mode=True)
    assert pp is not None

def test_show_3d_shape_image_silent():
    lungs = imgio.load_images(image_path_1, convert_to_grey=True, keep_3d_shape=True)

    pp = exp.show_image(lungs[1], silent_mode=True)
    assert pp is not None




import sys
import os
import pytest
import pandas as pd 
import arcus.ml.images.io as ami
import logging
import shutil

image_path = 'tests/resources/images/lungs'
image_url = 'https://cdn.sstatic.net/Sites/stackoverflow/img/apple-touch-icon@2.png'
image_url2 = 'https://github.com/arcus-azure/arcus/raw/master/media/arcus.png'
cache_directory = 'tests/resources/temp/io'
cache_file_name = 'testimage.png'

def setup_module(module):
    ''' Setup for the entire module '''
    # Do the actual setup stuff here
    pass

def setup_function(func):
    ''' Setup for test functions '''
    for filename in os.listdir(cache_directory):
        if (filename != '__emptyfile__'):     
            file_path = os.path.join(cache_directory, filename)
            try:
                if os.path.isfile(file_path) or os.path.islink(file_path):
                    os.unlink(file_path)
                #elif os.path.isdir(file_path):
                #    shutil.rmtree(file_path)
            except Exception as e:
                print('Failed to delete %s. Reason: %s' % (file_path, e))
    # cache_directory = os.path.join(os.getcwd(), 'tests/resources/temp/io')
    #if os.path.exists(cache_directory):
    #    shutil.rmtree(cache_directory)
    #os.mkdir(cache_directory)
    #os.chdir(cache_directory)
    #os.mkdir(cache_directory)

def test_load_single_image():
    image_file_name = 'tests/resources/images/lungs/CHNCXR_0001_0.png'
    image = ami.load_image_from_disk(image_file_name, image_size=40)

    assert(image.shape == (40, 40, 3))

def test_load_single_image_notexist():
    image_file_name = 'tests/resources/images/lungs/missingfile.png'
    with pytest.raises(FileNotFoundError) as fnf_err:
        image = ami.load_image_from_disk(image_file_name, image_size=40)
    assert image_file_name in str(fnf_err)

def test_load_images_defaults():
    image_list = ami.load_images(image_path)

    assert len(image_list) == 11

def test_load_images_max_number():
    image_list = ami.load_images(image_path, max_images=4)

    assert len(image_list) == 4

def test_load_images_resize_square():
    image_list = ami.load_images(image_path, image_size=40, max_images=1)

    assert len(image_list) == 1
    assert image_list[0].shape == (40, 40, 3)

def test_load_images_resize_tuple():
    image_list = ami.load_images(image_path, image_size=(40, 30), max_images=1)

    assert len(image_list) == 1
    assert image_list[0].shape == (40, 30, 3)

def test_load_images_grey():
    image_list = ami.load_images(image_path, image_size=40, max_images=1, convert_to_grey=True)

    assert image_list[0].shape == (40, 40)
    assert len(image_list) == 1

def test_load_images_grey_keep_shape():
    image_list = ami.load_images(image_path, image_size=40, max_images=1, convert_to_grey=True, keep_3d_shape=True)

    assert image_list[0].shape == (40, 40, 1)
    assert len(image_list) == 1

def test_load_images_grey_dont_keep_shape():
    image_list = ami.load_images(image_path, image_size=40, max_images=1, convert_to_grey=True, keep_3d_shape=False)

    assert image_list[0].shape == (40, 40)
    assert len(image_list) == 1

def test_load_images_extensions():
    image_list = ami.load_images(image_path, valid_extensions=['.gif'])

    assert len(image_list) == 0

def test_image_url_to_memory():
    image = ami.load_image_from_url(image_url)
    assert len(image.shape) == 3


def test_image_url_resize_memory():
    img_path = 'https://charting.vwdservices.com/tchart/tchart.aspx?user=Tijdnet&issue=360017012&layout=gradient-v1&startdate=3Y&enddate=today&res=endofday&width=876&height=400&format=image/png&culture=nl-BE'
    image = ami.load_image_from_url(img_path)
    assert len(image.shape) == 3

def test_image_url_to_cache():
    cached_file_name = os.path.join(cache_directory, cache_file_name)
    assert os.path.exists(cached_file_name)==False
    image = ami.load_image_from_url(image_url, cache_location=cache_directory, file_name=cache_file_name, force_download=False)
    assert len(image.shape) == 3
    assert os.path.exists(cached_file_name)
    disk_image = ami.load_image_from_disk(cached_file_name)
    assert disk_image.shape == image.shape

def test_image_url_through_cache():
    cached_file_name = os.path.join(cache_directory, cache_file_name)
    assert os.path.exists(cached_file_name)==False
    # Force download of 1st image
    image = ami.load_image_from_url(image_url, cache_location=cache_directory, file_name=cache_file_name, force_download=True)
    assert len(image.shape) == 3
    assert os.path.exists(cached_file_name)

    # Cached download of 2nd image
    image2 = ami.load_image_from_url(image_url2, cache_location=cache_directory, file_name=cache_file_name)
    assert len(image.shape) == 3
    assert os.path.exists(cached_file_name)
    # Even tough url was different, the image should be the same
    assert image.shape == image2.shape

def test_image_url_through_cache_missingfile():
    # Trigger missing 
    with pytest.raises(ValueError) as val_err:
        ami.load_image_from_url(image_url, cache_location=cache_directory, file_name='')
    assert 'file_name' in str(val_err.value)

def test_image_url_unauthorized():
    # Trigger missing 
    with pytest.raises(Exception) as val_err:
        ami.load_image_from_url('https://stockcharts.com/c-sc/sc?s=AAPL&p=W&yr=3&mn=0&dy=0&i=t5753201539c&r=1588670112897')
    assert '403' in str(val_err.value)

def test_image_url_headers():
    # Trigger missing 
    img = ami.load_image_from_url('https://stockcharts.com/c-sc/sc?s=AAPL&p=W&yr=3&mn=0&dy=0&i=t5753201539c&r=1588670112897',
        http_headers={'User-Agent': 'Postman'})
    assert len(img.shape) == 3

def test_load_images_df_multi():
    df = pd.read_csv('tests/resources/datasets/lung-files.csv')
    X, outputs = ami.load_images_from_dataframe(df, 'lungfile', ['result','maskfile'])
    assert len(X) == 11
    output1 = outputs[0]
    assert output1[0] in ('0', '1')
    assert os.path.exists(output1[1])

def test_load_images_set_df():
    df = pd.read_csv('tests/resources/datasets/lung-files.csv')
    X, y = ami.load_images_from_dataframe(df, 'lungfile', 'maskfile', target_as_image=True)
    assert len(X) == 11
    assert len(y[0].shape) == 3

def test_load_images_df():
    df = pd.read_csv('tests/resources/datasets/lung-files.csv')
    X, y = ami.load_images_from_dataframe(df, 'lungfile', 'result')
    assert len(X) == 11
    assert y[3] in (0, 1)

def test_load_images_df_sourcemissing():
    df = pd.read_csv('tests/resources/datasets/lung-files-missing-files.csv')
    # Two records contain a non existing file.  
    # 1 with source missing: output = 2
    # Another with destination missing: output = 0
    X, y = ami.load_images_from_dataframe(df, 'lungfile', 'result')
    assert len(X) == 10
    assert y[3] in (0, 1)

    assert 0 in y
    assert 1 in y
    assert not 2 in y

def test_load_images_df_destmissing():
    df = pd.read_csv('tests/resources/datasets/lung-files-missing-files.csv')
    # Two records contain a non existing file.  
    # 1 with source missing: output = 2
    # Another with destination missing: output = 0
    X, y = ami.load_images_from_dataframe(df, 'lungfile', 'maskfile', target_as_image=True)
    assert len(X) == 9



'''
The io module provides standard helper functions to load images from disk
'''

import logging
import os
import numpy as np
from cv2 import imread, imdecode, IMREAD_COLOR
import requests
from skimage import transform
from arcus.ml.images import conversion
from arcus.ml import dataframes as adf
from io import BytesIO
import pandas as pd

_logger = logging.getLogger()

def load_image_from_disk(path: str, image_size = None, 
                convert_to_grey: bool = False, keep_3d_shape = False) -> np.array:
    '''
    Loads an image from file, applying preformatting
    Args:
        path (str): The filename of the image to load
        image_size (tuple): The image size can be passed as tuple (W, H) or as int (W=H)
        convert_to_grey (bool): This would reduce the size (and shape) of the image in making it a greyscale
        keep_3d_shape (bool): Only used when convert_to_grey is true.  Will keep the images in shape (H,W,1) in that case
    Returns: 
        np.array: A numpy array that represents the image
    '''
    im = imread(path) 
    if im is None: # File does not exist
        raise FileNotFoundError(path)
    return conversion.prepare(im, image_size, convert_to_grey, keep_3d_shape)

def load_images(path: str, image_size = None, max_images: int = -1, 
                valid_extensions: np.array = ['.jpg','.jpeg','.gif','.png'],
                convert_to_grey: bool = False, keep_3d_shape = False) -> np.array:
    '''
    Loads the images from a specific folder
    Args:
        path (str): The path or folder name to load images from.  This can be a relative or fully qualified path
        image_size (tuple): The image size can be passed as tuple (W, H) or as int (W=H)
        max_images (int): The maximum amount of images to load from the folder.  If 0 or smaller, all images will be returned
        valid_extensions (np.array): The file extensions that should be filtered.  Defaults to jpg, jpeg, gif and png
        convert_to_grey (bool): This would reduce the size (and shape) of the image in making it a greyscale
        keep_3d_shape (bool): Only used when convert_to_grey is true.  Will keep the images in shape (H,W,1) in that case
    Returns: 
        np.array: A numpy array that contains all selected images represented as np.array
    '''
    images = []
    for f in os.listdir(path)[:]:
        ext = os.path.splitext(f)[1]
        if ext.lower() not in valid_extensions:
            continue
        
        im = load_image_from_disk(os.path.join(path,f), image_size, convert_to_grey, keep_3d_shape)

        images.append(im)
        if(len(images) >= max_images and max_images > 0):
            break
    
    return np.array(images)

def load_image_from_url(image_url: str, http_headers: dict = None, image_size = None, 
                convert_to_grey: bool = False, keep_3d_shape = False,
                cache_location: str = None, file_name: str = None, force_download: bool = False) -> np.array:
    '''
    Loads an image from a given url, applying preformatting and supporting file caching
    Args:
        image_url (str): The url to download the image.
        http_headers (dict): The http headers to pass with the request as a dictionary
        image_size (tuple): The image size can be passed as tuple (W, H) or as int (W=H)
        convert_to_grey (bool): This would reduce the size (and shape) of the image in making it a greyscale
        keep_3d_shape (bool): Only used when convert_to_grey is true.  Will keep the images in shape (H,W,1) in that case
        cache_location (str): When provided, the image will be cached to this folder location on disk
        file_name (str): The file name of the image to be cached
        force_download (bool): When true, the image will always be redownloaded and not retrieved from cache
    Returns: 
        np.array: A numpy array that represents the image
    '''
    image = []
    # If the users wants to cache, we first check if the image is 
    if (cache_location != None):
        if not os.path.exists(cache_location):
            os.makedirs(cache_location)
        if not file_name:
            raise ValueError('Since caching was asked, a file_name attribute should be passed too')

        image_path = os.path.join(cache_location, file_name)

        if os.path.exists(image_path) and force_download == False:
            # We can return the cached image
            _logger.debug('We are taking the cached image from %s and will not download from %s', image_path, image_url)

        else:
            _logger.info('Downloading image from %s and saving to %s', image_url, image_path)
            with requests.get(image_url, headers = http_headers, stream=True) as _http_response:
                if(str(_http_response.status_code).startswith('2')):
                    with open(image_path, 'wb') as fd:
                        for chunk in _http_response.iter_content(chunk_size=128):
                            fd.write(chunk)
                else:
                    raise Exception('HTTP error: ' + str(_http_response.status_code))
        # Returning image that has been persisted anyhow
        return load_image_from_disk(image_path, image_size, convert_to_grey, keep_3d_shape)
        
    else:
        _logger.info('Downloading image from %s and return in memory', image_url)
        with requests.get(image_url, headers = http_headers) as _http_response:
            if(str(_http_response.status_code).startswith('2')):
                image = np.asarray(bytearray(_http_response.content), dtype="uint8")
                image = imdecode(image, IMREAD_COLOR)
                return conversion.prepare(image, image_size, convert_to_grey, keep_3d_shape)
            else:
                raise Exception('HTTP error: ' + str(_http_response.status_code))

def load_images_from_dataframe(df: pd.DataFrame, image_column_name:str, target_column_name: str, 
                                image_size = None, max_images: int = -1, target_as_image: bool = False,
                                convert_to_grey: bool = False, keep_3d_shape = False) -> (np.array, np.array):
    '''
    Loads a set images from disk, based on the file name in a Data Frame.  
    And returns a related array (target) that can contain values from another column, or also images from disk

    Args:
        df (pd.DataFrame): the DataFrame, containing the references to the image
        image_column_name (str): The name of the column that contains the image reference
        target_column_name (str): The name of the column that contains the related target data
        image_size (tuple): The image size can be passed as tuple (W, H) or as int (W=H)
        max_images (int): The maximum amount of images to load from the folder.  If 0 or smaller, all images will be returned
        target_as_image (bool): Defines if the target column contains file names that should be loaded as image.  If not, the column data will be used in the target array
        convert_to_grey (bool): This would reduce the size (and shape) of the image in making it a greyscale
        keep_3d_shape (bool): Only used when convert_to_grey is true.  Will keep the images in shape (H,W,1) in that case
    
    Returns: 
        a tuple with the following objects:
        np.array: A numpy array that contains all selected images represented as np.array
        np.array: A numpy array that represents all targets that were asked
    '''
    images = []
    targets = []
    df = adf.shuffle(df)
    for idx, row in df.iterrows() :
        file = row[image_column_name]

        filecheck_ok = True
        try:
            im = load_image_from_disk(file, image_size, convert_to_grey, keep_3d_shape)
            if(target_as_image):
                target_file = row[target_column_name]
                if target_file:
                    target_im = load_image_from_disk(target_file, image_size, convert_to_grey, keep_3d_shape)
        except FileNotFoundError:
            _logger.warning('File ' + file + ' not found')
            filecheck_ok = False
        except:
            _logger.warning('File ' + file + ' resulted in an error')
            filecheck_ok = False

        if filecheck_ok :
            images.append(im)
            if target_as_image:
                # Check if the target file exists
                targets.append(target_im)
            else:
                if(isinstance(target_column_name, list)):
                    _current_targets = list()
                    for target_column in target_column_name:
                        _current_targets.append(row[target_column])
                    targets.append(_current_targets)
                else:
                    targets.append(row[target_column_name])
        else:
            _logger.warning('File ' + file + ' not found')

        if(len(images) >= max_images and max_images > 0):
            break
    
    return np.array(images), np.array(targets)